import inspect
import os

from werkzeug.routing import Map, Rule
from werkzeug.exceptions import HTTPException, NotFound, MethodNotAllowed
from werkzeug.middleware.shared_data import SharedDataMiddleware
from werkzeug.serving import run_simple

from .middleware import BaseMiddleware
from .response import TextResponse


class App():
    ROOT_DIR = None
    config = {
        'debug': True,
        'export_dirs': {
            '/static': 'static',
            '/media': 'media',
        },
        'host': '127.0.0.1',
        'port': 8000,
    }

    def __init__(self, filename, production=False):
        self.routes = {}
        self.url_rules = Map()
        self.middleware = BaseMiddleware(self)
        self.set_config(filename, production)

    def set_config(self, filename, production):
        if self.ROOT_DIR is None:
            self.ROOT_DIR = os.path.dirname(filename)
        if production:
            self.config['debug'] = False

    def set_export_dirs(self, export_dirs: dict):
        export_dirs = export_dirs.copy()
        for url, dir in export_dirs.items():
            export_dirs[url] = os.path.join(self.ROOT_DIR, dir)
        self.config['export_dirs'] = export_dirs

    def __call__(self, environ, start_response):
        return self.middleware(environ, start_response)

    def add_middleware(self, middleware_cls):
        if isinstance(self.middleware, SharedDataMiddleware):
            raise Exception("You can't add middleware after serving files!")
        self.middleware.add(middleware_cls)

    def dispatch(self, request):
        adapter = self.url_rules.bind_to_environ(request)
        try:
            endpoint, kwargs = adapter.match()
            handler = self.routes[endpoint]
            if inspect.isclass(handler):
                method = request.method
                handler = getattr(handler(), method.lower(), None)
                if handler is None:
                    return self.method_not_exist(request, method, endpoint)
            return handler(request, **kwargs)
        except NotFound:
            return self.error_404(request)
        except MethodNotAllowed:
            return self.error_405(request)
        except HTTPException as error:
            return error

    def error_404(self, request):
        return TextResponse('Page Not Found', status=404)

    def error_405(self, request):
        return TextResponse(
            f'Method `{request.method}` is not allowed!',
            status=405,
        )

    def method_not_exist(self, request, method, view_name):
        return TextResponse(
            "Server Error:\n"
            f"The `{view_name}` view does not support"
            f" the `{method}` method.",
            status=500,
        )

    def check_route_exists(self, path, endpoint):
        if endpoint in self.routes:
            return True
        for rule in self.url_rules.iter_rules():
            if rule.rule == path:
                return True
        return False

    def route(self, path, methods=None):
        def wrapper(handler):
            endpoint = handler.__name__
            assert not self.check_route_exists(path, endpoint), (
                "Path already exists!"
            )
            rule = Rule(path, endpoint=endpoint, methods=methods)
            self.routes[endpoint] = handler
            self.url_rules.add(rule)
            return handler
        return wrapper

    def serve_files(self, use_defaults=True, export_dirs=None):
        if self.config.get('debug', False) is False:
            raise Exception("You can't serve files in production mode!")
        if export_dirs is None:
            export_dirs = {}
        if use_defaults:
            export_dirs = self.config.get('export_dirs', {})
        self.set_export_dirs(export_dirs)
        self.middleware = SharedDataMiddleware(
            self.middleware,
            exports=self.config['export_dirs'],
        )

    def run(self, host=None, port=None, debug=True, reloader=True):
        if self.config.get('debug', False) is False:
            raise Exception(
                "You should use a web server in production mode!"
            )
        host = host or self.config.get('host', '127.0.0.1')
        port = port or self.config.get('port', 8000)
        run_simple(
            host,
            port,
            self.middleware,
            use_debugger=True,
            use_reloader=reloader,
        )



import os

from werkzeug.wrappers import Response
from jinja2 import Environment, FileSystemLoader


class TextResponse(Response):
    pass


class Render(Response):
    def __init__(
        self,
        template_name,
        template_dir="templates",
        context=None,
        **kwargs,
    ):
        if context is None:
            context = {}
        self.templates_env = Environment(
            loader=FileSystemLoader(os.path.abspath(template_dir))
        )
        text = self.templates_env.get_template(template_name).render(**context)
        super().__init__(text, content_type='text/html', **kwargs)



from .wsgi import App
from .response import TextResponse, Render
from .middleware import BaseMiddleware
from orm import DB, ResultConfig
from orm.operators import AND, OR, NOT
from orm import columns



from requests import Session
import pytest
from wsgiadapter import WSGIAdapter

from pirouz import (
    App, TextResponse, Render, BaseMiddleware
)

BASE_URL = 'http://testserver'


@pytest.fixture
def app():
    return App(__file__)


@pytest.fixture
def client(app):
    session = Session()
    session.mount(
        prefix=BASE_URL,
        adapter=WSGIAdapter(app),
    )
    return session


def test_basic_route(app, client):
    @app.route('/home')
    def home(request):
        return TextResponse('Home Page.')


def test_client_can_send_requests(app, client):
    TEXT = 'Hi Iam Mohammad Dori'

    @app.route('/hey/')
    def hey(request):
        return TextResponse(TEXT)
    assert client.get(f"{BASE_URL}/hey").text == TEXT
    assert client.get(f"{BASE_URL}/hey/").text == TEXT


def test_page_with_parameter(app, client):
    @app.route('/user/<int:id>/')
    def user(request, id):
        return TextResponse(f'Hi user with ID {id}')
    assert client.get(f"{BASE_URL}/user/1").text == 'Hi user with ID 1'
    assert client.get(f"{BASE_URL}/user/2/").text == 'Hi user with ID 2'


def test_404_error_response(app, client):
    assert client.get(f"{BASE_URL}/somethings/").status_code == 404



from __future__ import annotations
import os
import sqlite3
import inspect
from typing import Dict, List, NamedTuple, Tuple, Union
from dori_orm.operators import OPERATORS
from dori_orm.columns import ForeignKey


class GenerateTableName:
    def __get__(self, instance, owner) -> str:
        return owner.__name__.lower()


class GenerateDBName:
    def __get__(self, instance, owner) -> str:
        file_address = inspect.getfile(owner)
        db_name = os.path.basename(file_address)[:-3]
        return f"{db_name}.db"


class GetColumns:
    def get_columns(self, owner) -> Tuple[str, str]:
        class_variables: Dict[str, str] = owner.__dict__
        return (
            (key, value)
            for key, value in class_variables.items()
            # remove python magic method from class variables
            if not key.startswith('_')
        )

    def __get__(self, instance, owner) -> Dict[str, str]:
        columns = {
            name.lower(): f"{name.lower()} {value}"
            for name, value in self.get_columns(owner)
        }
        return {
            'id': 'id INTEGER PRIMARY KEY UNIQUE NOT NULL',
            **columns,
        }


class GetForeignKeys:
    def get_foreign_key_columns(self, owner) -> Tuple[str, str]:
        class_variables: Dict[str, str] = owner.__dict__
        return (
            (key, value.get_foreign_key())
            for key, value in class_variables.items()
            if isinstance(value, ForeignKey)
        )

    def __get__(self, instance, owner) -> Dict[str, str]:
        return {
            key: value
            for key, value in self.get_foreign_key_columns(owner)
        }


class Row:
    def __init__(self, db_name_: str, table_name_: str, **data):
        self.data = data
        for key, value in data.items():
            self.__setattr__(key, value)
        Row.db_name = db_name_
        Row.table_name = table_name_

    def remove(self):
        where = ' AND '.join([
            f'{key} = {repr(value)}'
            for key, value in self.data.items()
        ])
        query = f'DELETE FROM {self.table_name} WHERE {where}'
        self._execute(query)

    def update(self, **kwargs):
        where = ' AND '.join([
            f'{key} = {repr(value)}'
            for key, value in self.data.items()
        ])
        new_data = ', '.join([
            f'{key} = {repr(value)}'
            for key, value in kwargs.items()
        ])
        if new_data.strip():
            query = (
                f'UPDATE {self.table_name} SET {new_data} WHERE {where}'
            )
            self._execute(query)

    @ classmethod
    def _execute(cls, query: str):
        conn = sqlite3.connect(cls.db_name)
        conn.execute(query)
        conn.commit()
        conn.close()

    def __repr__(self) -> str:
        result = ', '.join([
            f'{repr(attr)}:{repr(value)}'
            for attr, value in self.data.items()
        ])
        return f"<{result}>"


class Rows:
    def __init__(self, rows: List[Row]):
        self.rows = rows

    def count(self) -> int:
        return len(self.rows)

    def first(self) -> Union[Row, None]:
        if self.rows:
            return self.rows[0]
        return None

    def last(self) -> Union[Row, None]:
        if self.rows:
            return self.rows[-1]
        return None

    def __repr__(self) -> str:
        return repr(self.rows)

    def __iter__(self) -> List[Row]:
        return iter(self.rows)


class ResultConfig(NamedTuple):
    limit: Union[int, None] = None
    order_by: Union[str, None] = None
    reverse: bool = False


class DB:
    db_name = GenerateDBName()
    table_name = GenerateTableName()
    columns = GetColumns()
    foreign_keys = GetForeignKeys()
    _query = ''

    def __init__(self, **data):
        data = {
            'id': self._get_max_id() + 1,
            **data,
        }
        for key, value in data.items():
            if key in self.foreign_keys.keys():
                data[key] = value.id
        self.data = data
        for key, value in data.items():
            self.__setattr__(key, value)
        self.id = self.id  # just for type hinting
        self.insert(**data)

    def __init_subclass__(cls, **kwargs):
        cls._manage_table()

    @classmethod
    def insert(cls, **data: dict):
        fields = ', '.join(data.keys())
        values = ', '.join(
            map(repr, data.values())
        )
        query = (f'INSERT OR IGNORE INTO {cls.table_name} '
                 f'({fields}) VALUES ({values});')
        cls._execute(query)

    @ classmethod
    def all(cls, config: Union[ResultConfig, None] = None) -> List[Row]:
        configs = cls._set_config(config)
        query = f'SELECT * FROM {cls.table_name}{configs};'
        return cls._fetchall(query)

    @ classmethod
    def get(cls, *fields: dict,
            config: Union[ResultConfig, None] = None) -> List[Row]:
        fields = [
            field
            for field in fields
            if field in cls.columns
        ]
        fields_string = ', '.join(fields) or '*'
        configs = cls._set_config(config)
        query = f'SELECT {fields_string} FROM {cls.table_name}{configs};'
        return cls._fetchall(query)

    @ classmethod
    def filter(cls, *args, config: Union[ResultConfig, None] = None,
               **kwargs) -> List[Row]:
        conditions = []
        for key, value in kwargs.items():
            if '__' in key:
                condition = cls._set_operator_filter(key, value)
            else:
                condition = f'{key} = {repr(value)}'
            if condition is not None:
                conditions.append(condition)
                condition = None
        conditions.extend(list(map(repr, args)))
        statements = ' AND '.join(conditions) or 'true'
        configs = cls._set_config(config)
        query = (
            f'SELECT * FROM {cls.table_name} WHERE {statements}{configs};'
        )
        return cls._fetchall(query)

    @ classmethod
    def max(cls, column_name: str):
        query = f'SELECT MAX({column_name}) FROM {cls.table_name};'
        result = cls._fetch_result(query)
        if result:
            return {column_name: result[0]}

    @ classmethod
    def min(cls, column_name: str):
        query = f'SELECT MIN({column_name}) FROM {cls.table_name};'
        result = cls._fetch_result(query)
        if result:
            return {column_name: result[0]}

    @ classmethod
    def avg(cls, column_name: str):
        query = f'SELECT AVG({column_name}) FROM {cls.table_name};'
        result = cls._fetch_result(query)
        if result:
            return {column_name: result[0]}

    @ classmethod
    def sum(cls, column_name: str):
        query = f'SELECT SUM({column_name}) FROM {cls.table_name};'
        result = cls._fetch_result(query)
        if result:
            return {column_name: result[0]}

    @ classmethod
    def count(cls):
        query = f'SELECT COUNT(1) FROM {cls.table_name};'
        result = cls._fetch_result(query)
        if result:
            return {'count': result[0]}

    @classmethod
    def first(cls) -> DB:
        query = f'SELECT * FROM {cls.table_name} WHERE id=1;'
        result = cls._fetch_result(query)
        if result is None:
            return None
        row = dict(zip(cls.columns.keys(), result))
        return Row(cls.db_name, cls.table_name, **row)

    @classmethod
    def last(cls) -> DB:
        max_id = cls._get_max_id()
        query = f'SELECT * FROM {cls.table_name} WHERE id={max_id};'
        result = cls._fetch_result(query)
        if result is None:
            return None
        row = dict(zip(cls.columns.keys(), result))
        return Row(cls.db_name, cls.table_name, **row)

    def remove(self):
        where = ' AND '.join([
            f'{key} = {repr(value)}'
            for key, value in self.data.items()
        ])
        query = f'DELETE FROM {self.table_name} WHERE {where};'
        self._execute(query)

    def update(self, **kwargs):
        where = ' AND '.join([
            f'{key} = {repr(value)}'
            for key, value in self.data.items()
        ])
        new_data = ', '.join([
            f'{key} = {repr(value)}'
            for key, value in kwargs.items()
        ])
        if new_data.strip():
            query = (
                f'UPDATE {self.table_name} SET {new_data} WHERE {where};'
            )
            self._execute(query)
        query = f'SELECT * FROM {self.table_name} WHERE id={self.id};'
        result = self._fetch_result(query)
        data = dict(zip(self.columns.keys(), result))
        self.data = data
        self.__dict__.update(data)

    @ classmethod
    def remove_table(cls):
        query = f'DROP TABLE {cls.table_name}'
        cls._execute(query)

    @ classmethod
    def queries(cls):
        return cls._query.strip()

    @ classmethod
    def _manage_table(cls) -> str:
        try:
            current_columns = cls._get_current_table_columns()
            if current_columns != list(cls.columns.keys()):
                cls._alter_columns(current_columns)
                cls._drop_columns(current_columns)
        except sqlite3.OperationalError:
            cls._create_table()

    @ classmethod
    def _create_table(cls) -> str:
        columns = list(cls.columns.values()).copy()
        foreign_keys = ', '.join(cls.foreign_keys.values())
        if foreign_keys:
            columns.append(foreign_keys)
        string_columns = ', '.join(columns)
        query = (f'CREATE TABLE IF NOT EXISTS {cls.table_name}'
                 f'({string_columns});')
        cls._execute(query)

    @classmethod
    def _alter_columns(cls, current_columns):
        new_columns = [
            value
            for column, value in cls.columns.items()
            if column not in current_columns
        ]
        for field in new_columns:
            query = f'ALTER TABLE {cls.table_name} ADD {field};'
            cls._execute(query)

    @classmethod
    def _drop_columns(cls, current_columns):
        removed_columns = [
            column
            for column in current_columns
            if column not in cls.columns.keys()
        ]
        for field in removed_columns:
            query = f'ALTER TABLE {cls.table_name} DROP {field};'
            cls._execute(query)

    @ staticmethod
    def _set_config(config: Union[ResultConfig, None]) -> str:
        if config is None:
            return ''
        limit, order_by, reverse = config
        if limit is None:
            limit = ''
        else:
            limit = f' LIMIT {limit}'
        if order_by is None:
            order_by = ''
            sorting = ''
        else:
            order_by = F'ORDER BY {order_by}'
            if reverse is False:
                sorting = ' ASC'
            else:
                sorting = ' DESC'
        return f' {order_by}{sorting}{limit}'

    @ staticmethod
    def _set_operator_filter(key: str, value: str):
        filter = key.split('__')
        if len(filter) != 2:
            return None
        key, operator = filter
        key, operator = key.lower(), operator.lower()
        if operator in OPERATORS:
            return (
                f'{key} {OPERATORS[operator]} {repr(value)}'
            )
        elif operator == 'between':
            start, *_, end = value
            return (
                f'{key} BETWEEN {repr(start)} AND {repr(end)}'
            )
        else:
            return None

    @ classmethod
    def _get_max_id(cls):
        query = f'SELECT MAX(id) FROM {cls.table_name}'
        result = cls._fetch_result(query)
        return result[0] or 0

    @ classmethod
    def _execute(cls, query: str):
        cls._query += f"{query}\n\n"
        conn = sqlite3.connect(cls.db_name)
        conn.execute(query)
        conn.commit()
        conn.close()

    @ classmethod
    def _fetchall(cls, query: str) -> Rows:
        cls._query += f"{query}\n\n"
        conn = sqlite3.connect(cls.db_name)
        cur = conn.cursor()
        cur.execute(query)
        rows = cur.fetchall()
        result = []
        for row in rows:
            row = dict(zip(
                cls.columns.keys(),
                row
            ))
            result.append(
                Row(cls.db_name, cls.table_name, **row)
            )
        conn.close()
        return Rows(result)

    @ classmethod
    def _fetch_result(cls, query: str):
        cls._query += f"{query}\n\n"
        conn = sqlite3.connect(cls.db_name)
        cur = conn.cursor()
        cur.execute(query)
        result = cur.fetchone()
        conn.close()
        return result

    @classmethod
    def _get_current_table_columns(cls):
        query = f'SELECT * FROM {cls.table_name}'
        conn = sqlite3.connect(cls.db_name)
        cur = conn.cursor()
        cur.execute(query)
        columns = [description[0] for description in cur.description]
        conn.close()
        return columns

    def __repr__(self) -> str:
        result = ', '.join([
            f'{repr(attr)}:{repr(value)}'
            for attr, value in self.data.items()
        ])
        return f"<{result}>"



OPERATORS = {
    'lt': '<',
    'lte': '<=',
    'gt': '>',
    'gte': '>=',
    'n': '!=',
    'in': 'IN',
    'like': 'LIKE',
}


class Operator:
    def __init__(self, *args, **kwargs):
        self.fields = kwargs
        self.args = args
        self.operator = self.__class__.__name__

    def generate_statements(self):
        operator = f' {self.operator} '
        statements = [
            f'{key}={repr(value)}'
            for key, value in self.fields.items()
        ]
        args_statements = operator.join(
            map(str, self.args))
        if args_statements:
            statements.append(args_statements)
        return f"({operator.join(statements)})"

    def __repr__(self) -> str:
        return self.generate_statements()


class AND(Operator):
    pass


class OR(Operator):
    pass


class NOT(Operator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.operator = 'AND'

    def __repr__(self) -> str:
        return f"NOT {super().__repr__()}"



from werkzeug.wrappers import Request


class BaseMiddleware:
    def __init__(self, app):
        self.app: BaseMiddleware = app

    def __call__(self, environ, start_response):
        request = Request(environ)
        response = self.app.dispatch(request)
        return response(environ, start_response)

    def add(self, middleware_cls):
        self.app = middleware_cls(self.app)

    def dispatch(self, request):
        request = self.process_request(request)
        response = self.app.dispatch(request)
        return self.process_response(request, response)

    def process_request(self, request):
        return request

    def process_response(self, request, response):
        return response



"""
python3 setup.py sdist
python -m twine upload dist/*
"""
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    requires = f.read().splitlines()
    print(requires)

setuptools.setup(
    name="pirouz",
    version="2.1.2",
    author="Mohammad Dori",
    author_email="mr.dori.dev@gmail.com",
    description="A web framework built with Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dori-dev/pirouz",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    install_requires=requires,
    python_requires=">=3.6",
)



from typing import Union


class Column:
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        self.type = None
        self.unique = unique
        self.nullable = nullable
        self.default = default

    def __repr__(self):
        constraints = ''
        if self.nullable:
            if self.default is not None:
                constraints += f'DEFAULT {self.default}'
        else:
            constraints += 'NOT NULL'
        if self.unique:
            constraints += ' UNIQUE'
        return f'{self.type} {constraints.strip()}'


class Int(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'INT'


class Integer(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'INTEGER'


class TinyInt(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'TINYINT'


class SmallInt(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'SMALLINT'


class MediumInt(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'MEDIUMINT'


class Text(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'TEXT'


class VarChar(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'VARCHAR(255)'


class Blob(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'BLOB'


class Real(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'REAL'


class Double(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'DOUBLE'


class Float(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'FLOAT'


class Numeric(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'NUMERIC'


class Decimal(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'DECIMAL(10,5)'


class Boolean(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'BOOLEAN'


class Date(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'DATE'


class DateTime(Column):
    def __init__(self, unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'DATETIME'


class ForeignKey(Column):
    def __set_name__(self, owner, name):
        self.column_name_ = name

    def __init__(self, reference: Union[object, str], unique: bool = False,
                 nullable: bool = True, default: str = None) -> None:
        super().__init__(unique, nullable, default)
        self.type = 'INTEGER'
        if isinstance(reference, object):
            reference = reference.__name__
        self.reference = reference.lower()

    def get_foreign_key(self):
        name = self.column_name_
        return f'FOREIGN KEY ({name}) REFERENCES {self.reference} (id)'
